@@ -1,5 +1,5 @@
 from functools import lru_cache
-from typing import Any, List, Tuple, Set, Dict, Union, MutableMapping
+from typing import Any, List, Tuple, Set, Dict, Union, MutableMapping, Iterator
 import textwrap
 
 import astroid
@@ -75,6 +75,48 @@ def is_map_field(fd):  # FIXME: too many selectors
     return is_composite(fd) and fd.message_type.has_options and fd.message_type.GetOptions().map_entry
 
 
+class TODO(object):
+    pass  # These fields are not assignable
+
+
+FIELD_TYPES = {
+    FieldDescriptor.TYPE_BOOL: bool,
+    FieldDescriptor.TYPE_BYTES: str,
+    FieldDescriptor.TYPE_DOUBLE: float,
+    FieldDescriptor.TYPE_ENUM: int,
+    FieldDescriptor.TYPE_FIXED32: float,
+    FieldDescriptor.TYPE_FIXED64: float,
+    FieldDescriptor.TYPE_FLOAT: float,
+    FieldDescriptor.TYPE_GROUP: TODO,
+    FieldDescriptor.TYPE_INT32: int,
+    FieldDescriptor.TYPE_INT64: int,
+    FieldDescriptor.TYPE_MESSAGE: TODO,
+    FieldDescriptor.TYPE_SFIXED32: float,
+    FieldDescriptor.TYPE_SFIXED64: float,
+    FieldDescriptor.TYPE_SINT32: int,
+    FieldDescriptor.TYPE_SINT64: int,
+    FieldDescriptor.TYPE_STRING: str,
+    FieldDescriptor.TYPE_UINT32: int,
+    FieldDescriptor.TYPE_UINT64: int,
+}
+
+def to_pytype(fd):
+    # type: (FieldDescriptor) -> type
+    if is_composite(fd):
+        return type(fd.message_type.name, (TODO,), {})  # XXX: such a hack!
+    return FIELD_TYPES[fd.type]
+
+def field_type_path(fd):
+    # type: (FieldDescriptor) -> Iterator[str]
+    if fd.containing_type is not None:
+        yield from field_type_path(fd.containing_type)
+    yield fd.name
+
+def full_name(fd):
+    # type: (FieldDescriptor) -> str
+    return '.'.join(field_type_path(fd))
+
+
 class SimpleDescriptor(object):
     def __init__(self, desc):
         # type: (Union[EnumDescriptor, Descriptor]) -> None
@@ -90,11 +132,19 @@ class SimpleDescriptor(object):
         # type: (FieldDescriptor) -> bool
         return fd.message_type.containing_type is self._desc
 
+    def is_typeof_field(self, fd):
+        # type: (FieldDescriptor) -> bool
+        return fd.message_type is self._desc
+
     @property
     def identifier(self):
         # type: () -> str
         return self._cls_hash
 
+    @property
+    def is_enum(self):
+        return self._is_protobuf_enum
+
     @property
     def name(self):
         # type: () -> str
@@ -136,6 +186,18 @@ class SimpleDescriptor(object):
         # type: () -> Dict[str, FieldDescriptor]
         return self._desc.fields_by_name
 
+    @property
+    def values(self):
+        # type: () -> Dict[str, int]
+        assert self._is_protobuf_enum, "Only makes sense for enum descriptors"
+        return {n: v.number for n, v in self._enum_desc.values_by_name.items()}
+
+    @property
+    def names(self):
+        # type: () -> Dict[int, str]
+        assert self._is_protobuf_enum, "Only makes sense for enum descriptors"
+        return {v.number: n for n, v in self._enum_desc.values_by_name.items()}
+
     @property
     def values_by_name(self):
         # type: () -> List[Tuple[str, int]]
@@ -168,7 +230,7 @@ class SimpleDescriptor(object):
     def external_fields(self):
         # type: () -> List[Tuple[str, str]]
         return [
-            (f.name, f.message_type.name) for f in self.message_fields
+            (f.name, full_name(f.message_type)) for f in self.message_fields
             if not self.is_nested(f)
         ]
 
@@ -263,15 +325,7 @@ def _template_message(desc, descriptor_registry):
         'self.{} = self.{}()  # inner_nonrepeated_fields'.format(field_name, field_type)
         for field_name, field_type in desc.inner_nonrepeated_fields
     ]
-    initialisers += [
-        'self.{} = {}()  # external_fields'.format(field_name, field_type)
-        for field_name, field_type in desc.external_fields
-    ]
 
-    # FIXME: some of these double up with the repeated fields
-    # e.g.
-    #   self.values = Inner()
-    #   self.values = InnerCompositeContainer()
 
     repeated_scalar_fields = [fd.name for fd in desc.fields if is_repeated(fd) and not is_composite(fd)]
     initialisers += [
@@ -279,16 +333,31 @@ def _template_message(desc, descriptor_registry):
         for field_name in repeated_scalar_fields
     ]
 
+    rcfields = {
+        fd for fd in desc.fields
+        if is_repeated(fd) and is_composite(fd) and not is_map_field(fd)
+    }
     repeated_composite_fields = [
         (fd.name, fd.message_type.name, desc.is_nested(fd))
-        for fd in desc.fields
-        if is_repeated(fd) and is_composite(fd) and not is_map_field(fd)
+        for fd in rcfields
     ]
     initialisers += [
         _template_composite_field(desc.name, field_name, field_type, is_nested)
         for field_name, field_type, is_nested in repeated_composite_fields
     ]
 
+    # TODO: refactor this
+    external_fields = [
+        (f.name, full_name(f.message_type)) for f in desc.message_fields
+        if not desc.is_nested(f)
+        if f not in rcfields  # don't want to double up above
+    ]
+    initialisers += [
+        'self.{} = {}()  # external_fields'.format(field_name, field_type)
+        for field_name, field_type in external_fields
+
+    ]
+
     args = ['self'] + ['{}=None'.format(f) for f in slots]
     init_str = 'def __init__({argspec}):\n{initialisers}\n'.format(
         argspec=', '.join(args),
@@ -391,6 +460,7 @@ def transform_module(mod):
         cls = mod_node_to_class(mod, name)
         try:
             for local_name, node in transform_descriptor_to_class(cls):
+                node.parent = mod
                 mod.locals[local_name] = [node]
         except NotImplementedError:
             pass
